******************************************************************************
* PROGRAM ----: KBDIALOG.PRG 
* AUTHOR -----: Kevin E. Saffer 
* COPYRIGHT --: None, placed into the public domain
* CREATED ----: 6/12/1994 at 17:50
******************************************************************************
#include "fivewin.ch"

* offsets into the browse color array aColors
#define KTB_HEADCLR      1  // header area color pair
#define KTB_DATACLR      2  // data area color pair
#define KTB_FOOTCLR      3  // footer area color pair
#define KTB_LINECLR      4  // line color pair
#define KTB_CELLCLR      5  // cell pointer color pair

* offsets into each browse color array element
#define KTB_FOREGROUND   1
#define KTB_BACKGROUND   2

* additional defines not in fivewin.ch
#define CS_DBLCLKS       8
#define WM_LBUTTONDBLCLK 515
#define WM_ERASEBKGND    20

STATIC lRegistered := .F.

******************************************************************************
* class definition, instance variables, and method prototypes
******************************************************************************
CLASS KbDialog FROM tControl

  * passed window coordinates
  DATA nTlr           // top row coordinate in pixels
  DATA nTlc           // top left column coordinate in pixels
  DATA nBrr           // bottom row coordinate in pixels
  DATA nBrc           // bottom right column coordinate in pixels

  * arrays 
  DATA aColors        // array of browse colors
  DATA aColumns       // array of column information objects
  DATA aReference     // array reference if browsing an array

  * column object array pointers
  DATA nTopColPtr     // offset of leftmost displayed column object
  DATA nCurColPtr     // offset of currently displayed column object
  DATA nBotColPtr     // offset of rightmost displayed column object
  DATA nFrzColPtr     // offset of the frozen column

  * grid position control variables
  DATA nGridType      // grid type
  DATA nGridHeight    // total height of each cell grid
  DATA nGridRowAdj    // row spacing between the text box and cell grid
  DATA nGridColAdj    // column spacing between the text box and cell grid
  DATA nGridRowOff    // row spacing between each cell
  DATA nGridColOff    // column spacing between each cell

  * text position control variables
  DATA nTextFntHgt    // height of the font as reported by windows
  DATA nTextBoxAdj    // additional text box height added to base font height
  DATA nTextColAdj    // column offset into the text box
  DATA nTextBoxPix    // pixel adjustement for grid types 1 and 2

  * screen row pointers
  DATA nHeadRowTop    // header area upper screen row
  DATA nHeadRowBot    // header area lower screen row
  DATA nHeadOffset    // spacing between the header and data area
  DATA nDataRowTop    // data area upper screen row
  DATA nDataRowBot    // data area lower screen row
  DATA nFootOffset    // spacing between the data and footer area
  DATA nFootRowTop    // footer area upper screen row
  DATA nFootRowBot    // footer area lower screen row
  DATA nCurRowPtr     // current display row in pixels
  DATA nBotRowPtr     // last used display row in pixels

  * source data element pointers
  DATA nDataTopElm    // topmost displayed data item
  DATA nDataCurElm    // currently displayed data item
  DATA nDataBotElm    // bottom displayed data item
  DATA nArrCurElm     // current element for array browsing

  * control variables
  DATA lClearPaint    // logical to clear window before next paint call
  DATA lRePaint       // logical to repaint areas during next paint call
  DATA lCalcFont      // logical to force font recalculation during next paint call
  DATA lCalcSize      // logical to force coordinate recalculation during next paint call
  DATA lCellPaint     // logical to determine single-cell or full-line painting
  DATA nLastKey       // numeric to indicate last key pressed

  * source data operation blocks
  DATA bBOF           // test beginning of file code block
  DATA bEOF           // test end of file code block
  DATA bGOTOP         // goto top code block
  DATA bGOBOTTOM      // goto bottom code block
  DATA bGOTO          // goto <n> code block
  DATA bSKIPPER       // skip code block
  DATA bRECNO         // retrieve data element code block
  DATA bRECCOUNT      // return total number of data items
  DATA bRECPOS        // return logical record position
  DATA bGOTOPOS       // goto logical record position

  * source database pointers for restoration
  DATA cAlias
  DATA nWorkArea
  DATA nIndexOrd

  * scroll bar objects
  DATA oVScroll       // vertical scrollbar 
  DATA oHScroll       // horizontal scrollbar

  METHOD AddColumn()
  METHOD Array()
  METHOD BlankPaint()
  METHOD CalcFont()
  METHOD CalcSize()
  METHOD DataPaint()
  METHOD Default()
  METHOD DeLight()
  METHOD FootPaint()
  METHOD GetDlgCode()
  METHOD GoBottom()
  METHOD GoDown()
  METHOD GoFarLeft()
  METHOD GoFarRight()
  METHOD GoLeft()
  METHOD GoLeftCol()
  METHOD GoPageDown()
  METHOD GoPageLeft()
  METHOD GoPageRight()
  METHOD GoPageUp()
  METHOD GoRight()
  METHOD GoRightCol()
  METHOD GoTop()
  METHOD GoUp()
  METHOD HandleEvent()
  METHOD HeadPaint()
  METHOD HiLight()
  METHOD HScroll()
  METHOD Init()
  METHOD KeyDown()
  METHOD LButtonDown()
  METHOD LinePaint()
  METHOD New() CONSTRUCTOR
  METHOD Paint()
  METHOD SetCellColor()
  METHOD SetDataColor()
  METHOD SetFootColor()
  METHOD SetFreeze()
  METHOD SetGrid()
  METHOD SetHeadColor()
  METHOD SetLineColor()
  METHOD SetNewFont()
  METHOD VScroll()

ENDCLASS

******************************************************************************
METHOD New( oWnd, nTlr, nTlc, nBrr, nBrc, oFont ) CLASS KbDialog
******************************************************************************

  DEFAULT oWnd := GetWndDefault(), ;
          nTlr := 0, ;
          nTlc := 0, ;
          nBrr := oWnd:nBottom - oWnd:nTop, ;
          nBrc := oWnd:nRight - oWnd:nLeft, ;
          oFont := TFont():New( "Fixedsys", 8, 12 )

  ::oWnd := oWnd
  ::nTop := nTlr / 2
  ::nLeft := nTlc / 2
  ::nBottom := nBrr / 2
  ::nRight := nBrc / 2
  ::nStyle := nOR( WS_CHILD, WS_VSCROLL, WS_HSCROLL, WS_BORDER, WS_VISIBLE, WS_TABSTOP )
  ::cCaption := ""
  ::oFont := oFont
  ::lFocused := .F.
  ::lDrag := .F.
  ::lCaptured := .F.

  ::aColors := { { CLR_WHITE, CLR_BLUE }, ;
                 { CLR_BLACK, CLR_LIGHTGRAY }, ;
                 { CLR_WHITE, CLR_GRAY }, ;
                 { CLR_BLACK, CLR_LIGHTGRAY }, ;
                 { CLR_WHITE, CLR_BLACK } ;
               }

  ::aColumns := {} 
  ::aReference := NIL 

  ::SetColor( ::aColors[ KTB_DATACLR, KTB_FOREGROUND ], ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )

  * set up our virtual screen coordinates, adjusting for the scroll bars
  ::nTlr := nTlr  
  ::nTlc := nTlc
  ::nBrr := nBrr - 16
  ::nBrc := nBrc - 16

  * initialize the column array pointers
  ::nTopColPtr := 1
  ::nCurColPtr := 1
  ::nBotColPtr := 1
  ::nFrzColPtr := 0

  * default to grid type 3 (3D Boxes)
  ::SetGrid( 3 )

  * initialize the screen and data pointers
  ::nHeadOffset := 0
  ::nFootOffset := 0
  ::nCurRowPtr := 0
  ::nBotRowPtr := 0 
  ::nDataTopElm := 0
  ::nDataCurElm := 0
  ::nDataBotElm := 0
  ::nArrCurElm := 1

  * set up lastkey
  ::nLastKey := 0

  * create standard database file operation blocks
  ::bBOF := { || BOF() }
  ::bEOF := { || EOF() }
  ::bGOTOP  := { || DBGOTOP() }
  ::bGOBOTTOM := { || DBGOBOTTOM() }
  ::bGOTO := { | n | DBGOTO( n ) }
  ::bSKIPPER := { | n | DBSKIP( n ) }
  ::bRECNO := { || RECNO() }
  ::bRECCOUNT := { || RECCOUNT() }
  ::bRECPOS := { || IIF( INDEXORD() = 0, RECNO(), NTXPOS( INDEXORD(), RECNO() ) ) }
  ::bGOTOPOS := { | n | DBGOTO( n ) }

  * store the database alias and indexorder
  ::cAlias := ALIAS()
  ::nWorkArea := SELECT()
  ::nIndexOrd := INDEXORD()

  * create the scrollbars
  DEFINE SCROLLBAR ::oVScroll VERTICAL OF Self
  DEFINE SCROLLBAR ::oHScroll HORIZONTAL OF Self

  IF .NOT. lRegistered
    ::Register( nOR( CS_VREDRAW, CS_HREDRAW, CS_GLOBALCLASS, CS_DBLCLKS ) )
    lRegistered := .T.
  ENDIF

  IF oWnd:lVisible
    ::Create()
    ::Default()
    ::lVisible := .T.
  ELSE
    oWnd:DefControl( Self )
    ::lVisible := .F.
  ENDIF

RETURN 0

******************************************************************************
METHOD Default() CLASS KbDialog
******************************************************************************

  * set up to recalculate the fonts and grids, the screen will already be clear
  ::lRePaint := .T.
  ::lCalcFont := .T.
  ::lCalcSize := .T.
  ::lClearPaint := .F.

  * range the scrollbars
  ::oVScroll:SetRange( 1, Eval( ::bRECCOUNT ) )
  ::oHScroll:SetRange( 1, Len( ::aColumns ) )

RETURN 0

******************************************************************************
METHOD CalcSize() CLASS KbDialog
******************************************************************************
  LOCAL nRow         // screen row counter
  LOCAL nHeadRowMax  // maximum number of header rows
  LOCAL nFootRowMax  // maximum number of footer rows
  LOCAL nStrPos      // pointer into a header/footer title string
  LOCAL nColPtr      // pointer into the column object array
  LOCAL cString      // isolated title string for justification

  * reset the switch
  ::lCalcSize := .F.

  * reset the data display area to the full window
  ::nDataRowTop := ::nTlr
  ::nDataRowBot := ::nBrr

  * initially assume that no headers or footers are present
  ::nHeadRowTop := -1
  ::nHeadRowBot := -1
  ::nFootRowTop := -1
  ::nFootRowBot := -1

  * setup default maximum number of header/footer rows
  nHeadRowMax := 0
  nFootRowMax := 0

  * loop through each column object
  FOR nColPtr := 1 TO LEN( ::aColumns ) 

    * check for a header
    IF ::aColumns[ nColPtr ]:cHeader <> NIL

      * assume this column's header is one cell high
      nRow := 1

      * check for multiline headers and count up how many are present
      IF ";" $ ::aColumns[ nColPtr ]:cHeader
        FOR nStrPos := 1 TO LEN( ::aColumns[ nColPtr ]:cHeader )
          IF SUBSTR( ::aColumns[ nColPtr ]:cHeader, nStrPos, 1 ) = ";"
            nRow++
          ENDIF
        NEXT
      ENDIF

      * check to assign the largest header row count to the grand total
      IF nRow > nHeadRowMax
        nHeadRowMax := nRow
      ENDIF
  
    ENDIF

    * check for a footer
    IF ::aColumns[ nColPtr ]:cFooter <> NIL

      * assume this column's footer is one cell high
      nRow := 1

      * check for multiline footers and count up how many are present
      IF ";" $ ::aColumns[ nColPtr ]:cFooter
        FOR nStrPos := 1 TO LEN( ::aColumns[ nColPtr ]:cFooter )
          IF SUBSTR( ::aColumns[ nColPtr ]:cFooter, nStrPos, 1 ) = ";"
            nRow++
          ENDIF
        NEXT
      ENDIF

      * check to assign the largest footer row count to the grand total
      IF nRow > nFootRowMax
        nFootRowMax := nRow
      ENDIF

    ENDIF

  NEXT

  * if headers are present, assign header row coords and adjust the data area
  IF nHeadRowMax > 0
    ::nHeadRowTop := ::nDataRowTop
    ::nHeadRowBot := ::nHeadRowTop + ::nGridRowAdj + ( nHeadRowMax * ( ::nTextFntHgt + ::nTextBoxAdj ) ) + ::nGridRowAdj  
    ::nDataRowTop := ::nHeadRowBot + ::nGridRowOff + ::nHeadOffset
  ENDIF

  * if footers are present, assign footer row coords and adjust the data area
  IF nFootRowMax > 0
    ::nFootRowBot := ::nDataRowBot
    ::nFootRowTop := ::nFootRowBot - ( ::nGridRowAdj + ( nFootRowMax * ( ::nTextFntHgt + ::nTextBoxAdj ) ) + ::nGridRowAdj )
    ::nDataRowBot := ::nFootRowTop - ::nHeadOffset
  ENDIF

RETURN 0

******************************************************************************
METHOD HandleEvent( nMsg, nWParam, nLParam ) CLASS KbDialog
******************************************************************************

  * check for database browsing
  IF ::aReference = NIL

    * check for a workarea or index order switch
    IF SELECT() <> ::nWorkArea .OR. INDEXORD() <> ::nIndexOrd

      * restore the selected workarea and index order
      SELECT (::nWorkArea)
      SET ORDER TO ::nIndexOrd

    ENDIF

    * check for a different database in this area
    IF ALIAS() <> ::cAlias
      ::End()
      RETURN 0
    ENDIF

    * check for record pointer movement
    IF ::nDataCurElm <> 0 .AND. ::nDataCurElm <> Eval( ::bRECNO )

      * reposition on the current data element
      Eval( ::bGOTO, ::nDataCurElm )

    ENDIF

  ENDIF

  DO CASE
  CASE nMsg == WM_GETDLGCODE
    RETURN ::GetDlgCode()

  CASE nMsg == WM_KEYDOWN
    RETURN ::KeyDown( nWParam, nLParam )

  CASE nMsg == WM_LBUTTONDOWN
    RETURN ::LButtonDown( nHiWord( nLParam ), nLoWord( nLParam ), nWParam )

  CASE nMsg == WM_VSCROLL
    RETURN ::VScroll( nWParam, nLParam )

  CASE nMsg == WM_HSCROLL
    RETURN ::HScroll( nWParam, nLParam )

  CASE nMsg == WM_SIZE
    ::lCalcSize := .T.
    ::lClearPaint := .T.

  CASE nMsg == WM_PAINT
    ::BeginPaint()
    ::Paint()
    ::EndPaint()
    RETURN 0

  ENDCASE

RETURN Super:HandleEvent( nMsg, nWParam, nLParam )

*******************************************************************************
METHOD KeyDown( nKey, nFlags ) CLASS KbDialog
*******************************************************************************
  LOCAL bKeyAction := SetKey( nKey )

  IF ::bKeyDown != NIL 
    Eval( ::bKeyDown, nKey, nFlags )
  ENDIF

  DO CASE

  CASE nKey == VK_TAB
    ::oWnd:GoNextCtrl( ::hWnd )
    RETURN 0

  CASE nKey == VK_F4 .and. GetKeyState( VK_CONTROL ) .AND. ;
     ::oWnd != NIL .and. ::oWnd:ChildLevel( TMdiChild() ) > 0
     ::oWnd:KeyDown( nKey, nFlags )
    RETURN 0

  CASE nKey == VK_F6 .and. GetKeyState( VK_CONTROL ) .AND. ;
     ::oWnd != nil .and. ::oWnd:ChildLevel( TMdiChild() ) > 0
     ::oWnd:KeyDown( nKey, nFlags )
    RETURN 0

  CASE nKey == VK_F1
    HelpTopic( ::nHelpId )
    RETURN 0

  CASE bKeyAction != nil
    Eval( bKeyAction, ProcName( 4 ), ProcLine( 4 ) )
    RETURN 0

  CASE nKey == VK_RIGHT .AND. GetKeyState( VK_CONTROL )  
    RETURN ::GoPageRight()

  CASE nKey == VK_RIGHT
    RETURN ::GoRight()

  CASE nKey == VK_LEFT .AND. GetKeyState( VK_CONTROL )
    RETURN ::GoPageLeft()

  CASE nKey == VK_LEFT
    RETURN ::GoLeft()

  CASE nKey == VK_HOME .AND. GetKeyState( VK_CONTROL )
    RETURN ::GoFarLeft()

  CASE nKey == VK_HOME
    RETURN ::GoLeftCol()

  CASE nKey == VK_END .AND. GetKeyState( VK_CONTROL )
    RETURN ::GoFarRight()

  CASE nKey == VK_END
    RETURN ::GoRightCol()

  CASE nKey == VK_UP
    RETURN ::GoUp()

  CASE nKey == VK_DOWN
    RETURN ::GoDown()

  CASE nKey == VK_PRIOR .AND. GetKeyState( VK_CONTROL )
    RETURN ::GoTop()

  CASE nKey == VK_NEXT .AND. GetKeyState( VK_CONTROL )
    RETURN ::GoBottom()

  CASE nKey == VK_PRIOR
    RETURN ::GoPageUp()

  CASE nKey == VK_NEXT
    RETURN ::GoPageDown()

  ENDCASE

RETURN 0

******************************************************************************
METHOD Init( hWnd ) CLASS KbDialog
******************************************************************************

  Super:Init( hWnd ) 
  ::Default()

RETURN 0

******************************************************************************
METHOD Array( aReference ) CLASS KbDialog
******************************************************************************
  
  * store the passed reference to the array into the browser
  ::aReference := aReference

  * initialize the internal array element pointer
  ::nArrCurElm := 1

  * alter the data movement blocks to operate on an array
  ::bBOF := { || ::nArrCurElm < 1 }
  ::bEOF := { || ::nArrCurElm > Len( ::aReference ) }
  ::bGOTOP  := { || ::nArrCurElm := 1 }
  ::bGOBOTTOM := { || ::nArrCurElm := Len( ::aReference ) }
  ::bGOTO := { | n | ::nArrCurElm := n }
  ::bSKIPPER := { | n | ::nArrCurElm += n }
  ::bRECNO := { || ::nArrCurElm }
  ::bRECCOUNT := { || Len( ::aReference )  }
  ::bRECPOS := { || ::nArrCurElm }
  ::bGOTOPOS := { | n | ::nArrCurElm := n }

RETURN 0

******************************************************************************
METHOD Paint() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * see if we are iconized
  IF IsIconic( ::hWnd )
    ::SendMsg( WM_ERASEBKGND, ::hDC, 0 )
    DrawIcon( ::hDC, 0, 0, IIF( ::oIcon != NIL, ::oIcon:hIcon, ExtractIcon( "user.exe" ) ) )
    RETURN 0
  ENDIF

  * check to bypass area paint routines
  IF .NOT. ::lRePaint
    ::lRePaint := .T.
    RETURN 0
  ENDIF

  * check to perform internal font specific recalculations
  IF ::lCalcFont
    ::CalcFont()
  ENDIF

  * check to perform internal coordinate recalculations
  IF ::lCalcSize
    ::CalcSize()
  ENDIF

  * check identical line/cell colors to switch to cell painting (much faster)
  ::lCellPaint := .F.
  IF ::aColors[ KTB_LINECLR, KTB_FOREGROUND ] = ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] .AND. ;
     ::aColors[ KTB_LINECLR, KTB_BACKGROUND ] = ::aColors[ KTB_DATACLR, KTB_BACKGROUND ]
    ::lCellPaint := .T.
  ENDIF

  * obtain a display context handle and set the current font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * check to clear the browse window
  IF ::lClearPaint
    SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
    SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )
    EXTTEXTOUT( hDC, ::nTlr, ::nTlc, { ::nTlr, ::nTlc, ::nBrr + 2, ::nBrc + 2 }, "" )
    ::lClearPaint := .F.
  ENDIF

  * check for browse headers and paint 'em
  IF ::nHeadRowTop >= 0
    ::HeadPaint( hDC )
  ENDIF

  * paint the data display, specifying cell grids
  ::DataPaint( hDC, .T. )

  * check for browse footers and paint 'em
  IF ::nFootRowTop >= 0
    ::FootPaint( hDC )
  ENDIF

  * check to update the column scrollbar
  IF ::oHScroll:GetPos() <> ::nCurColPtr
    ::oHScroll:SetPos( ::nCurColPtr )
  ENDIF

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD GetDlgCode() CLASS KbDialog
******************************************************************************
RETURN nOR( DLGC_WANTARROWS, DLGC_WANTCHARS )

******************************************************************************
METHOD AddColumn( oColumn ) CLASS KbDialog
******************************************************************************

  Aadd( ::aColumns, oColumn )
  ::lCalcFont := .T.
  ::lCalcSize := .T.

  * set the horizontal scrollbar range
  ::oHScroll:SetRange( 1, Len( ::aColumns ) )

RETURN 0

******************************************************************************
METHOD CalcFont() CLASS KbDialog
******************************************************************************
  LOCAL nRow         // screen row counter
  LOCAL nStrPos      // pointer into a header/footer title string
  LOCAL nColPtr      // pointer into the column object array
  LOCAL cString      // isolated title string for justification
  LOCAL nDataLength  // length of the cell data
  LOCAL nLength      // length of the header/footer
  LOCAL nMaxLength   // test for deta length override

  * reset the switch
  ::lCalcFont := .F.

  * select our font
  SelectObject( ::hDC, ::oFont:hFont )

  * obtain the font height and establish the grid cell height
  ::nTextFntHgt := ABS( ::oFont:nHeight )
  ::nTextBoxAdj := Int(::nTextFntHgt / 5) + 2
  ::nGridHeight := ::nGridRowAdj + ::nTextFntHgt + ::nTextBoxAdj + ::nGridRowAdj

  * loop through each column object
  FOR nColPtr := 1 TO LEN( ::aColumns ) 

    * check to see if we should calculate this cell width
    IF ::aColumns[ nColPtr ]:nProgWidth = NIL

      * obtain the block result
      cString := Eval(::aColumns[ nColPtr ]:bData) 

      * check for a character return value and determine the best cell width
      IF VALTYPE(cString) = "C"

        * assume the length of the data return string
        nDataLength := LEN( cString )
        nMaxLength := 0

        * check for a header
        IF ::aColumns[ nColPtr ]:cHeader <> NIL

          * check the header for a wider string
          nLength := 0
          FOR nStrPos := 1 TO LEN( ::aColumns[ nColPtr ]:cHeader )
  
            * check for multi-line header
            IF SUBSTR( ::aColumns[ nColPtr ]:cHeader, nStrPos, 1 ) = ";"
              IF nLength > nMaxLength
                nMaxLength := nLength
              ENDIF
              nLength := 0
              nStrPos++
              LOOP
            ENDIF
  
            nLength++

          NEXT
  
          * check to update the maximum header/footer length
          IF nLength > nMaxLength
            nMaxLength := nLength
          ENDIF

        ENDIF

        * check for a footer
        IF ::aColumns[ nColPtr ]:cFooter <> NIL

          * check the footer for a wider string
          nLength := 0
          FOR nStrPos := 1 TO LEN( ::aColumns[ nColPtr ]:cFooter )
  
            * check for multi-line footer
            IF SUBSTR( ::aColumns[ nColPtr ]:cFooter, nStrPos, 1 ) = ";"
              IF nLength > nMaxLength
                nMaxLength := nLength
              ENDIF
              nLength := 0
              nStrPos++
              LOOP
            ENDIF
  
            nLength++

          NEXT

          * check to update the header/footer length
          IF nLength > nMaxLength
            nMaxLength := nLength
          ENDIF
  
        ENDIF

        * check to update the data length
        IF nMaxLength > nDataLength
          nDataLength := nMaxLength
        ENDIF

        * store the cell width into the column object
        ::aColumns[ nColPtr ]:nCellWidth := GetTextWidth( ::hDC, Replicate("B", nDataLength ) ) + ( ::nTextColAdj * 2 ) + ( ::nGridColAdj * 2 )

      ELSE

        * error, truncate the column width to show the programmer
        ::aColumns[ nColPtr ]:nCellWidth := 10

      ENDIF

    ELSE

      * activate the programmer specified cell width
      ::aColumns[ nColPtr ]:nCellWidth := ::aColumns[ nColPtr ]:nProgWidth   

    ENDIF

  NEXT

  * reset the current row pointer
  ::nCurRowPtr := 0

RETURN 0

******************************************************************************
METHOD HeadPaint( hDC ) CLASS KbDialog
******************************************************************************
  LOCAL cChar        // header title string character
  LOCAL cString      // isolated title string for justification
  LOCAL nCol         // screen column position in pixels
  LOCAL nColPtr      // pointer into the column object array
  LOCAL nRow         // screen row position in pixels
  LOCAL nStrPos      // pointer into the header title string
  LOCAL lReleaseDC   // determines if a display context was created internally

  * check for a passed display context, obtain it if needed
  IF hDC = NIL
    lReleaseDC := .T.
    hDC := GetDC( ::hWnd )
    SelectObject( hDC, ::oFont:hFont )
  ELSE
    lReleaseDC := .F.
  ENDIF

  * set the display context to the header colors
  SetTextColor( hDC, ::aColors[ KTB_HEADCLR, KTB_FOREGROUND ] )
  SetBkColor( hDC, ::aColors[ KTB_HEADCLR, KTB_BACKGROUND ] )

  * assign leftmost screen column position
  nCol := ::nTlc

  * loop through each displayable column
  nColPtr = IIF( ::nFrzColPtr = 0, ::nTopColPtr, 1 )
  FOR nColPtr := nColPtr TO LEN(::aColumns) 

    * check for frozen columns
    IF ::nFrzColPtr > 0 .AND. nColPtr = ::nFrzColPtr + 1
      nColPtr = ::nTopColPtr
    ENDIF

    * draw the cell grid
    IF ::nGridType = 3
      WNDBOXRAISED( hDC, ::nHeadRowTop, nCol, ::nHeadRowBot, nCol + ::aColumns[ nColPtr ]:nCellWidth )
    ELSE
      WNDBOX( hDC, ::nHeadRowTop, nCol, ::nHeadRowBot, nCol + ::aColumns[ nColPtr ]:nCellWidth )
    ENDIF

    * header text display loop
    nStrPos := 1
    FOR nRow := ::nHeadRowTop + ::nGridRowAdj TO ::nHeadRowBot - ::nGridRowAdj - 1 STEP ::nTextFntHgt + ::nTextBoxAdj

      * extract next text segment from the header
      cString := ""
      FOR nStrPos := nStrPos TO LEN( ::aColumns[ nColPtr ]:cHeader )
        cChar := SUBSTR( ::aColumns[ nColPtr ]:cHeader, nStrPos, 1 )
        IF cChar = ";"
          nStrPos++
          EXIT
        ENDIF
        cString += cChar
      NEXT

      * justify the header text
      IF ::aColumns[ nColPtr ]:nJustify = 1
        cString := KBWPADC( hDC, cString, ::aColumns[ nColPtr ]:nCellWidth - ( ::nGridColAdj * 2 ) - ( ::nTextColAdj * 2 ) )
      ELSEIF ::aColumns[ nColPtr ]:nJustify = 2
        cString := KBWPADL( hDC, cString, ::aColumns[ nColPtr ]:nCellWidth - ( ::nGridColAdj * 2 ) - ( ::nTextColAdj * 2 ) )
      ENDIF

      * display the header text
      EXTTEXTOUT( hDC, nRow, nCol + ::nGridColAdj + ::nTextColAdj, ;
        { nRow + ::nTextBoxPix, nCol + ::nGridColAdj + ::nTextBoxPix, nRow + ::nTextFntHgt + ::nTextBoxAdj, nCol + ::aColumns[ nColPtr ]:nCellWidth - ::nGridColAdj }, ;
        cString )

    NEXT

    * increment the screen column by this cell's width
    nCol += ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff

    * see if we've gone past the rightmost screen column and exit if so
    IF nCol > ::nBrc
      EXIT
    ENDIF

  NEXT

  * check to release the display context to Windows
  IF lReleaseDC
    ReleaseDC( ::hWnd, hDC )
  ENDIF

RETURN 0

******************************************************************************
METHOD DataPaint( hDC, lDrawGrids ) CLASS KbDialog
******************************************************************************
  LOCAL nRow         // screen row position in pixels
  LOCAL nCol         // screen column position in pixels
  LOCAL nColPtr      // pointer into the column object array
  LOCAL nRowStop     // allows one last row if no footers are present
  LOCAL lReleaseDC   // determines if a display context was created internally
  LOCAL lSetCurCol   // determines if we need to set the current column pointer

  * check for a passed display context, obtain it if needed
  IF hDC = NIL
    lReleaseDC := .T.
    hDC := GetDC( ::hWnd )
    SelectObject( hDC, ::oFont:hFont )
  ELSE
    lReleaseDC := .F.
  ENDIF

  * check for the drawgrid flag, assume none if not passed
  IF lDrawGrids = NIL
    lDrawGrids := .F.
  ENDIF

  * set the display context to the data colors
  SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
  SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )

  * check to allow the display of one last cell beyond the bottom row
  IF ::nFootRowTop = -1
    nRowStop := 0
  ELSE
    nRowStop := ::nGridHeight
  ENDIF

  * check for valid source data position
  IF Eval( ::bEOF )
    Eval( ::bGOBOTTOM )
    IF Eval( ::bEOF )
      Eval( ::bGOTOP )
    ENDIF
  ENDIF

  * check for operating mode
  IF Eval( ::bEOF )
    * we are browsing an empty source, reset all
    ::nDataTopElm := 0
    ::nDataCurElm := 0
    ::nDataBotElm := 0
    ::nCurRowPtr := ::nDataRowTop
    ::nBotRowPtr := ::nDataRowTop
  ELSEIF ::nDataCurElm = 0
    * first time startup, reset source data and starting screen pointers
    ::nDataTopElm := Eval( ::bRECNO )
    ::nDataCurElm := Eval( ::bRECNO )
    ::nDataBotElm := Eval( ::bRECNO )
    ::nCurRowPtr := ::nDataRowTop
    ::nBotRowPtr := ::nDataRowTop
  ENDIF

  * check for a valid screen row pointer, set to top data row if needed
  IF ::nCurRowPtr < ::nDataRowTop
    ::nCurRowPtr := ::nDataRowTop
  ENDIF

  nRow := ::nDataRowTop

  * we paint data with a valid current data element
  IF ::nDataCurElm <> 0

    * position to the top displayed source data element
    Eval( ::bGOTO, ::nDataTopElm )
    ::nDataCurElm := -1

    * loop through each data row
    FOR nRow := ::nDataRowTop TO ::nDataRowBot - nRowStop STEP ::nGridHeight + ::nGridRowOff

      * check for end of data and exit if so
      IF Eval( ::bEOF )
        Eval( ::bGOBOTTOM )
        EXIT
      ENDIF 

      * paint the current data line
      ::LinePaint( hDC, nRow, lDrawGrids, .T. )

      * check to see if we're on the current row
      IF nRow = ::nCurRowPtr
 
        * record current source data element
        ::nDataCurElm := Eval( ::bRECNO )

        * check for a valid current column array pointer and adjust if needed
        lSetCurCol := .F.
        IF ::nCurColPtr < ::nTopColPtr .OR. ::nCurColPtr > ::nBotColPtr 
          lSetCurCol := .T.
        ENDIF
        IF ::nCurColPtr > 0 .AND. ::nCurColPtr <= ::nFrzColPtr
          lSetCurCol := .F.
        ENDIF

        IF lSetCurCol
          ::nCurColPtr := ::nTopColPtr

          * check to highlight the current cell or paint the entire line
          IF ::lCellPaint
            ::HiLight( hDC )
          ELSE
            ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
          ENDIF

          SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
          SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )

        ENDIF

      ENDIF

      * check to update the bottom source data element and screen row
      IF nRow + ::nGridHeight + ::nGridRowOff < ::nDataRowBot - nRowStop

        ::nDataBotElm := Eval( ::bRECNO )
        ::nBotRowPtr := nRow

      ENDIF

      * move onto the next record
      Eval( ::bSKIPPER, 1 )
  
    NEXT

    * check to see if the current source data element was found
    IF ::nDataCurElm > 0

      * repositon on the current source data element
      Eval( ::bGOTO, ::nDataCurElm )

    ELSE

      * use the last source data element displayed on the screen
      ::nDataCurElm := ::nDataBotElm
      ::nCurRowPtr := ::nBotRowPtr

      * position to the current source data element and hilight it
      Eval( ::bGOTO, ::nDataCurElm )

      * check to highlight the current cell or paint the entire line
      IF ::lCellPaint
        ::HiLight( hDC )
      ELSE
        ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
      ENDIF

      SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
      SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )

    ENDIF

    * update the scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )
    ::oHScroll:SetPos( ::nCurColPtr )

  ENDIF

  * check to paint blank text cells the rest of the way down the screen
  FOR nRow := nRow TO ::nDataRowBot - nRowStop STEP ::nGridHeight + ::nGridRowOff

    * paint the current blank data line
    ::BlankPaint( hDC, nRow, lDrawGrids )

  NEXT

  * check for a valid data element
  IF ::nDataCurElm = 0
    ::HiLight( hDC )
  ENDIF

  * check to release the display context to Windows
  IF lReleaseDC
    ReleaseDC( ::hWnd, hDC )
  ENDIF

RETURN 0

******************************************************************************
METHOD LinePaint( hDC, nRow, lDrawGrids, lSetColors ) CLASS KbDialog
******************************************************************************
  LOCAL nCol
  LOCAL nColPtr

  * assign leftmost screen column position
  nCol := ::nTlc

  IF lSetColors
    * check to see if we're on the current screen row
    IF nRow = ::nCurRowPtr
      * switch to the line color
      SetTextColor( hDC, ::aColors[ KTB_LINECLR, KTB_FOREGROUND ] )
      SetBkColor( hDC, ::aColors[ KTB_LINECLR, KTB_BACKGROUND ] )
    ENDIF
  ELSE
    * switch to the data color
    SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
    SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )
  ENDIF

  * loop through each displayable column
  nColPtr = IIF( ::nFrzColPtr = 0, ::nTopColPtr, 1 )
  FOR nColPtr := nColPtr TO LEN(::aColumns) 

    * check for frozen columns
    IF ::nFrzColPtr > 0 .AND. nColPtr = ::nFrzColPtr + 1
      nColPtr = ::nTopColPtr
    ENDIF

    * check to see if we should draw grids
    IF lDrawGrids
      IF ::nGridType = 3
        WNDBOXRAISED( hDC, nRow, nCol, nRow + ::nGridHeight, nCol + ::aColumns[ nColPtr ]:nCellWidth )
      ELSE
        WNDBOX( hDC, nRow, nCol, nRow + ::nGridHeight, nCol + ::aColumns[ nColPtr ]:nCellWidth )
      ENDIF
    ENDIF
    
    * check to see if we're on the current screen row and column object
    IF nRow = ::nCurRowPtr .AND. nColPtr = ::nCurColPtr .AND. lSetColors

      SetTextColor( hDC, ::aColors[ KTB_CELLCLR, KTB_FOREGROUND ] )
      SetBkColor( hDC, ::aColors[ KTB_CELLCLR, KTB_BACKGROUND ] )

    ENDIF

    * display the column data retrieval block
    EXTTEXTOUT( hDC, nRow + ::nGridRowAdj, nCol + ::nGridColAdj + ::nTextColAdj, ;
      { nRow + ::nGridRowAdj + ::nTextBoxPix, nCol + ::nGridColAdj + ::nTextBoxPix, nRow + ::nGridRowAdj + ::nTextFntHgt + ::nTextBoxAdj, nCol + ::aColumns[ nColPtr ]:nCellWidth - ::nGridColAdj }, ;
      Eval(::aColumns[ nColPtr ]:bData) )
    
    * check to see if we're on the current screen row and column object
    IF nRow = ::nCurRowPtr .AND. nColPtr = ::nCurColPtr .AND. lSetColors

      SetTextColor( hDC, ::aColors[ KTB_LINECLR, KTB_FOREGROUND ] )
      SetBkColor( hDC, ::aColors[ KTB_LINECLR, KTB_BACKGROUND ] )

    ENDIF

    * update this column's screen position
    ::aColumns[ nColPtr ]:nScreenCol := nCol
    
    * increment the screen column by this cell's width
    nCol += ::aColumns[ nColPtr ]:nCellWidth + IIF( nColPtr = LEN( ::aColumns ), 0, ::nGridColOff )
    
    * see if we've gone past the rightmost screen column and exit if so
    IF nCol > ::nBrc
      IF nColPtr = ::nTopColPtr .OR. nColPtr < ::nFrzColPtr
        ::nBotColPtr := nColPtr
      ENDIF
      EXIT
    ENDIF
    
    * update the rightmost displayed column
    ::nBotColPtr := nColPtr

  NEXT

  * check to see if we're on the current screen row
  IF nRow = ::nCurRowPtr .AND. lSetColors
    SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
    SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )
  ENDIF

RETURN 0

******************************************************************************
METHOD BlankPaint( hDC, nRow, lDrawGrids ) CLASS KbDialog
******************************************************************************
  LOCAL nCol
  LOCAL nColPtr

  * assign leftmost screen column position
  nCol := ::nTlc

  * loop through each displayable column
  nColPtr = IIF( ::nFrzColPtr = 0, ::nTopColPtr, 1 )
  FOR nColPtr := nColPtr TO LEN(::aColumns) 

    * check for frozen columns
    IF ::nFrzColPtr > 0 .AND. nColPtr = ::nFrzColPtr + 1
      nColPtr = ::nTopColPtr
    ENDIF

    * check to see if we should draw grids
    IF lDrawGrids
      IF ::nGridType = 3
        WNDBOXRAISED( hDC, nRow, nCol, nRow + ::nGridHeight, nCol + ::aColumns[ nColPtr ]:nCellWidth )
      ELSE
        WNDBOX( hDC, nRow, nCol, nRow + ::nGridHeight, nCol + ::aColumns[ nColPtr ]:nCellWidth )
      ENDIF
    ENDIF
    
    * display a blank data retrieval block
    EXTTEXTOUT( hDC, nRow + ::nGridRowAdj, nCol + ::nGridColAdj + ::nTextColAdj, ;
      { nRow + ::nGridRowAdj + ::nTextBoxPix, nCol + ::nGridColAdj + ::nTextBoxPix, nRow + ::nGridRowAdj + ::nTextFntHgt + ::nTextBoxAdj, nCol + ::aColumns[ nColPtr ]:nCellWidth - ::nGridColAdj }, ;
      "" )
    
    * update this columns screen position
    ::aColumns[ nColPtr ]:nScreenCol := nCol
    
    * increment the screen column by this cell's width
    nCol += ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
    
    * see if we've gone past the rightmost screen column and exit if so
    IF nCol > ::nBrc
      EXIT
    ENDIF
    
    * update the rightmost displayed column
    ::nBotColPtr := nColPtr
    
  NEXT

RETURN 0

******************************************************************************
METHOD FootPaint( hDC ) CLASS KbDialog
******************************************************************************
  LOCAL cChar        // footer title string character
  LOCAL cString      // isolated title string for justification
  LOCAL nCol         // screen column position in pixels
  LOCAL nColPtr      // pointer into the column object array
  LOCAL nRow         // screen row position in pixels
  LOCAL nStrPos      // pointer into the footer title string
  LOCAL lReleaseDC   // determines if a display context was created internally

  * check for a passed display context, obtain it if needed
  IF hDC = NIL
    lReleaseDC := .T.
    hDC := GetDC( ::hWnd )
    SelectObject( hDC, ::oFont:hFont )
  ELSE
    lReleaseDC := .F.
  ENDIF

  * set the display context to the footer colors
  SetTextColor( hDC, ::aColors[ KTB_FOOTCLR, KTB_FOREGROUND ] )
  SetBkColor( hDC, ::aColors[ KTB_FOOTCLR, KTB_BACKGROUND ] )

  * assign leftmost screen column position
  nCol := ::nTlc

  * loop through each displayable column
  nColPtr = IIF( ::nFrzColPtr = 0, ::nTopColPtr, 1 )
  FOR nColPtr := nColPtr TO LEN(::aColumns) 

    * check for frozen columns
    IF ::nFrzColPtr > 0 .AND. nColPtr = ::nFrzColPtr + 1
      nColPtr = ::nTopColPtr
    ENDIF

    * draw the cell grid
    IF ::nGridType = 3
      WNDBOXRAISED( hDC, ::nFootRowTop, nCol, ::nFootRowBot, nCol + ::aColumns[ nColPtr ]:nCellWidth )
    ELSE
      WNDBOX( hDC, ::nFootRowTop, nCol, ::nFootRowBot, nCol + ::aColumns[ nColPtr ]:nCellWidth )
    ENDIF

    * footer text display loop
    nStrPos := 1
    FOR nRow := ::nFootRowTop + ::nGridRowAdj TO ::nFootRowBot - ::nGridRowAdj - 1 STEP ::nTextFntHgt + ::nTextBoxAdj

      * extract next text segment from the footer
      cString := ""
      FOR nStrPos := nStrPos TO LEN( ::aColumns[ nColPtr ]:cFooter )
        cChar := SUBSTR( ::aColumns[ nColPtr ]:cFooter, nStrPos, 1 )
        IF cChar = ";"
          nStrPos++
          EXIT
        ENDIF
        cString += cChar
      NEXT

      * justify the footer text
      IF ::aColumns[ nColPtr ]:nJustify = 1
        cString := KBWPADC( hDC, cString, ::aColumns[ nColPtr ]:nCellWidth - ( ::nGridColAdj * 2 ) - ( ::nTextColAdj * 2 ) )
      ELSEIF ::aColumns[ nColPtr ]:nJustify = 2
        cString := KBWPADL( hDC, cString, ::aColumns[ nColPtr ]:nCellWidth - ( ::nGridColAdj * 2 ) - ( ::nTextColAdj * 2 ) )
      ENDIF

      * display the footer text
      EXTTEXTOUT( hDC, nRow, nCol + ::nGridColAdj + ::nTextColAdj, ;
        { nRow + ::nTextBoxPix, nCol + ::nGridColAdj + ::nTextBoxPix, nRow + ::nTextFntHgt + ::nTextBoxAdj, nCol + ::aColumns[ nColPtr ]:nCellWidth - ::nGridColAdj }, ;
        cString )

    NEXT

    * increment the screen column by this cell's width
    nCol += ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff

    * see if we've gone past the rightmost screen column and exit if so
    IF nCol > ::nBrc
      EXIT
    ENDIF

  NEXT

  * check to release the display context to Windows
  IF lReleaseDC
    ReleaseDC( ::hWnd, hDC )
  ENDIF

RETURN 0

******************************************************************************
METHOD HiLight( hDC ) CLASS KbDialog
******************************************************************************
  LOCAL lReleaseDC   // determines if a display context was created internally
  LOCAL nCol         // screen display column position 
   
  * check for a passed display context, obtain it if needed
  IF hDC = NIL
    lReleaseDC := .T.
    hDC := GetDC( ::hWnd )
    SelectObject( hDC, ::oFont:hFont )
  ELSE
    lReleaseDC := .F.
  ENDIF

  * set the display context to the cell colors
  SetTextColor( hDC, ::aColors[ KTB_CELLCLR, KTB_FOREGROUND ] )
  SetBkColor( hDC, ::aColors[ KTB_CELLCLR, KTB_BACKGROUND ] )

  * obtain the screen column position from the current column object
  nCol := ::aColumns[ ::nCurColPtr ]:nScreenCol

  * display the column data retrieval block
  EXTTEXTOUT( hDC, ::nCurRowPtr + ::nGridRowAdj, nCol + ::nGridColAdj + ::nTextColAdj, ;
    { ::nCurRowPtr + ::nGridRowAdj + ::nTextBoxPix, nCol + ::nGridColAdj + ::nTextBoxPix, ::nCurRowPtr + ::nGridRowAdj + ::nTextFntHgt + ::nTextBoxAdj, nCol + ::aColumns[ ::nCurColPtr ]:nCellWidth - ::nGridColAdj }, ;
    Eval(::aColumns[ ::nCurColPtr ]:bData) )

  * check to release the display context to Windows
  IF lReleaseDC
    ReleaseDC( ::hWnd, hDC )
  ENDIF

RETURN 0

******************************************************************************
METHOD DeLight( hDC ) CLASS KbDialog
******************************************************************************
  LOCAL lReleaseDC   // determines if a display context was created internally
  LOCAL nCol         // screen display column position 
   
  * check for a passed display context, obtain it if needed
  IF hDC = NIL
    lReleaseDC := .T.
    hDC := GetDC( ::hWnd )
    SelectObject( hDC, ::oFont:hFont )
  ELSE
    lReleaseDC := .F.
  ENDIF

  * set the display context to the line colors
  SetTextColor( hDC, ::aColors[ KTB_LINECLR, KTB_FOREGROUND ] )
  SetBkColor( hDC, ::aColors[ KTB_LINECLR, KTB_BACKGROUND ] )
  
  * obtain the screen column position from the current column object
  nCol := ::aColumns[ ::nCurColPtr ]:nScreenCol
    
  * display the column data retrieval block
  EXTTEXTOUT( hDC, ::nCurRowPtr + ::nGridRowAdj, nCol + ::nGridColAdj + ::nTextColAdj, ;
    { ::nCurRowPtr + ::nGridRowAdj + ::nTextBoxPix, nCol + ::nGridColAdj + ::nTextBoxPix, ::nCurRowPtr + ::nGridRowAdj + ::nTextFntHgt + ::nTextBoxAdj, nCol + ::aColumns[ ::nCurColPtr ]:nCellWidth - ::nGridColAdj }, ;
    Eval(::aColumns[ ::nCurColPtr ]:bData) )
 
  * check to release the display context to Windows
  IF lReleaseDC
    ReleaseDC( ::hWnd, hDC )
  ENDIF

RETURN 0

******************************************************************************
METHOD GoRight() CLASS KbDialog
******************************************************************************
  LOCAL nColPtr
  LOCAL nCol

  * check to see if the cell pointer is in the last column element
  IF ::nCurColPtr = LEN( ::aColumns )
    RETURN 0
  ENDIF

  * delight the current cell
  ::DeLight()

  ::nCurColPtr++
  * check for frozen columns
  IF ::nFrzColPtr > 0 .AND. ::nCurColPtr = ::nFrzColPtr + 1
    ::nCurColPtr := ::nTopColPtr
  ENDIF

  * update the scrollbar
  ::oHScroll:SetPos( ::nCurColPtr )

  * check to see if the current column is on screen, highlight and return if so
  IF ::nCurColPtr <= ::nBotColPtr
    ::HiLight()
    RETURN 0
  ENDIF

  * check to see if there was only one column on screen, repaint if so
  IF ::nTopColPtr = ::nBotColPtr
    ::nTopColPtr := ::nCurColPtr
    ::lClearPaint := .T.
    ::BeginPaint()
    ::Paint()
    ::EndPaint()
    RETURN 0
  ENDIF

  * adjust the top column element pointer until the new column moves onscreen
  nCol := ::aColumns[ ::nBotColPtr ]:nScreenCol + ;
          ::aColumns[ ::nBotColPtr ]:nCellWidth + ::nGridColOff + ;
          ::aColumns[ ::nCurColPtr ]:nCellWidth + ::nGridColOff

  FOR nColPtr := ::nTopColPtr TO ::nCurColPtr
    nCol -= ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
    IF nCol <= ::nBrc
      nColPtr++
      EXIT
    ENDIF
  NEXT
  ::nTopColPtr := nColPtr

  * for now, clear and paint the entire screen
  ::lClearPaint := .T.
  ::BeginPaint()
  ::Paint()
  ::EndPaint()
  RETURN 0

  * this scrolling routine is not visually appealing, and is not used just yet

  * scroll the data area left and repaint
  ScrollWindow( ::hWnd, -( ::aColumns[ ::nTopColPtr ]:nScreenCol - ::nTlc ), 0, ;
                  {  ::nDataRowTop, ::nTlc, ::nDataRowBot, ::nBrc + 100 }, 0 )

  ::BeginPaint()
  ::Paint()
  ::EndPaint()

RETURN 0

******************************************************************************
METHOD GoLeft() CLASS KbDialog
******************************************************************************

  * check to see if the cell pointer is in the first column element
  IF ::nCurColPtr = 1
    RETURN 0
  ENDIF

  * delight the current cell
  ::DeLight()

  ::nCurColPtr--
  * update the scrollbar
  ::oHScroll:SetPos( ::nCurColPtr )

  * check to see if the current column is on screen, highlight and return if so
  IF ::nCurColPtr >= ::nTopColPtr .OR. ::nCurColPtr <= ::nFrzColPtr
    ::HiLight()
    RETURN 0
  ENDIF

  * adjust the top column element pointer
  IF ::nCurColPtr > ::nFrzColPtr
    ::nTopColPtr := ::nCurColPtr
  ENDIF

  * for now, clear and paint the entire screen
  ::lClearPaint := .T.
  ::BeginPaint()
  ::Paint()
  ::EndPaint()
  RETURN 0

  * this scrolling routine is not visually appealing, and is not used just yet

  * scroll the data area right and repaint
  ScrollWindow( ::hWnd, ( ::aColumns[ ::nTopColPtr ]:nCellWidth + ::nGridColOff ), 0, ;
                  {  ::nTlr, ::nTlc, ::nBrr, ::nBrc }, 0 )

  ::BeginPaint()
  ::Paint()
  ::EndPaint()

RETURN 0

******************************************************************************
METHOD GoRightCol() CLASS KbDialog
******************************************************************************

  * check to see if the cell pointer is in the right column element
  IF ::nCurColPtr = ::nBotColPtr
    RETURN 0
  ENDIF

  * delight the current cell
  ::DeLight()

  * set the current column element pointer and hilight
  ::nCurColPtr := ::nBotColPtr
  ::Hilight()
  ::oHScroll:SetPos( ::nCurColPtr )

RETURN 0

******************************************************************************
METHOD GoLeftCol() CLASS KbDialog
******************************************************************************

  * check to see if the cell pointer is in the left column element
  IF ( ::nFrzColPtr = 0 .AND. ::nCurColPtr = ::nTopColPtr ) .OR. ;
     ( ::nFrzColPtr > 0 .AND. ::nCurColPtr = 1 )
    RETURN 0
  ENDIF

  * delight the current cell
  ::DeLight()

  * set the current column element pointer and hilight
  IF ::nFrzColPtr = 0
    ::nCurColPtr := ::nTopColPtr
  ELSE
    ::nCurColPtr := 1
  ENDIF

  ::Hilight()
  ::oHScroll:SetPos( ::nCurColPtr )

RETURN 0

******************************************************************************
METHOD GoPageRight() CLASS KbDialog
******************************************************************************

  * check to see if the cell pointer is in the last column element
  IF ::nCurColPtr = LEN( ::aColumns )
    RETURN 0
  ENDIF

  * determine if the last column is on the screen
  IF ::nBotColPtr = LEN( ::aColumns )
    * delight the current cell
    ::DeLight()

    * set the current column element pointer and hilight
    ::nCurColPtr := ::nBotColPtr
    ::Hilight()
    ::oHScroll:SetPos( ::nCurColPtr )
    RETURN 0
  ENDIF

  ::nTopColPtr := ::nBotColPtr + 1
  ::nCurColPtr := ::nBotColPtr + 1
  ::oHScroll:SetPos( ::nCurColPtr )
  ::lClearPaint := .T.
  ::BeginPaint()
  ::Paint()
  ::EndPaint()

RETURN 0

******************************************************************************
METHOD GoPageLeft() CLASS KbDialog
******************************************************************************
  LOCAL nColPtr
  LOCAL nCol

  * check to see if the cell pointer is in the first column element
  IF ::nCurColPtr = 1
    RETURN 0
  ENDIF

  * determine if column 1 is on screen
  IF ::nTopColPtr = 1 .OR. ( ::nFrzColPtr > 0 .AND. ::nTopColPtr = ::nFrzColPtr + 1 )
    * delight the current cell
    ::DeLight()

    * set the current column element pointer and hilight
    IF ::nCurColPtr > ::nTopColPtr
      ::nCurColPtr := ::nTopColPtr
    ELSE
      ::nCurColPtr := 1
    ENDIF
    ::Hilight()
    ::oHScroll:SetPos( ::nCurColPtr )
    RETURN 0
  ENDIF

  nColPtr := ::nTopColPtr - 1
  nCol := ::nBrc - ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
  DO WHILE nColPtr > 1
    nColPtr--
    IF nCol - ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff < ::nTlc
      nColPtr++
      EXIT
    ENDIF
    IF ::nFrzColPtr > 0 .AND. nCol - ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff < ::aColumns[ ::nFrzColPtr ]:nCellWidth + ::nGridColOff
      nColPtr++
      EXIT
    ENDIF
    nCol -= ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
  ENDDO

  IF ::nFrzColPtr > 0 .AND. nColPtr <= ::nFrzColPtr
    nColPtr := ::nFrzColPtr + 1
  ENDIF

  ::nTopColPtr := nColPtr
  ::nCurColPtr := nColPtr
  ::oHScroll:SetPos( ::nCurColPtr )
  ::lClearPaint := .T.
  ::BeginPaint()
  ::Paint()
  ::EndPaint()

RETURN 0

******************************************************************************
METHOD GoFarRight() CLASS KbDialog
******************************************************************************
  LOCAL nColPtr
  LOCAL nCol

  * check to see if the cell pointer is in the last column element
  IF ::nCurColPtr = LEN( ::aColumns )
    RETURN 0
  ENDIF

  * determine if the last column is on the screen
  IF ::nBotColPtr = LEN( ::aColumns )
    * delight the current cell
    ::DeLight()

    * set the current column element pointer and hilight
    ::nCurColPtr := ::nBotColPtr
    ::Hilight()
    ::oHScroll:SetPos( ::nCurColPtr )
    RETURN 0
  ENDIF

  nColPtr := LEN( ::aColumns )
  nCol := ::nBrc - ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
  DO WHILE nColPtr > 1
    nColPtr--
    IF nCol - ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff < ::nTlc
      nColPtr++
      EXIT
    ENDIF
    IF ::nFrzColPtr > 0 .AND. nCol - ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff < ::aColumns[ ::nFrzColPtr ]:nCellWidth + ::nGridColOff
      nColPtr++
      EXIT
    ENDIF
    nCol -= ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
  ENDDO

  ::nTopColPtr := nColPtr
  ::nCurColPtr := LEN( ::aColumns )
  ::oHScroll:SetPos( ::nCurColPtr )
  ::lClearPaint := .T.
  ::BeginPaint()
  ::Paint()
  ::EndPaint()

RETURN 0

******************************************************************************
METHOD GoFarLeft() CLASS KbDialog
******************************************************************************

  * check to see if the cell pointer is in the first column element
  IF ::nCurColPtr = 1
    RETURN 0
  ENDIF

  * determine if column 1 is on screen
  IF ::nTopColPtr = 1 .OR. ( ::nFrzColPtr > 0 .AND. ::nTopColPtr = ::nFrzColPtr + 1 )
    * delight the current cell
    ::DeLight()

    * set the current column element pointer and hilight
    ::nCurColPtr := 1
    ::Hilight()
    ::oHScroll:SetPos( ::nCurColPtr )
    RETURN 0
  ENDIF

  ::nTopColPtr := IIF( ::nFrzColPtr = 0, 1, ::nFrzColPtr + 1)
  ::nCurColPtr := 1
  ::oHScroll:SetPos( ::nCurColPtr )
  ::lClearPaint := .T.
  ::BeginPaint()
  ::Paint()
  ::EndPaint()

RETURN 0

******************************************************************************
METHOD GoUp() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * skip backward and check for bof, return if so
  Eval( ::bSKIPPER, -1 )
  IF Eval( ::bBOF )
    Eval( ::bGOTOP )
    RETURN 0
  ENDIF

  * obtain our display context and font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * skip forward and delight the previous row
  Eval( ::bSKIPPER, 1 )

  * check for line/cell color differences and either delight or paint the line
  IF ::aColors[ KTB_CELLCLR, KTB_FOREGROUND ] = ::aColors[ KTB_LINECLR, KTB_FOREGROUND ] .AND. ;
     ::aColors[ KTB_CELLCLR, KTB_BACKGROUND ] = ::aColors[ KTB_LINECLR, KTB_BACKGROUND ]
    ::DeLight()
  ELSE
    ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
  ENDIF

  * skip to the new source data element and record it
  Eval( ::bSKIPPER, -1 )
  ::nDataCurElm := Eval( ::bRECNO )

  * deduct the screen row pointer
  ::nCurRowPtr -= ::nGridHeight + ::nGridRowOff

  * check to see if we need to scroll the data area downwards
  IF ::nCurRowPtr < ::nDataRowTop

    * restore the top screen row
    ::nCurRowPtr := ::nDataRowTop

    * backup and record the top data element 
    Eval( ::bGOTO, ::nDataTopElm )
    Eval( ::bSKIPPER, -1 )
    ::nDataTopElm := Eval( ::bRECNO )

    * backup and record the bottom data element 
    Eval( ::bGOTO, ::nDataBotElm )
    Eval( ::bSKIPPER, -1 )
    ::nDataBotElm := Eval( ::bRECNO )

    * reposition on the current data element
    Eval( ::bGOTO, ::nDataCurElm )

    * inform our paint method to keep out of the way
    ::lRePaint := .F.

    * scroll the data area up and repaint to satisfy windows
    ::lRePaint := .F.
    ScrollWindow( ::hWnd, 0, ::nGridHeight + ::nGridRowOff, ;
                  {  ::nDataRowTop, ::nTlc, ::nDataRowBot, ::nBrc }, ;
                  {  ::nDataRowTop, ::nTlc, ::nDataRowBot, ::nBrc } )
    ::BeginPaint()
    ::Paint()
    ::EndPaint()

    * paint the data line with grids
    ::LinePaint( hDC, ::nCurRowPtr, .T., .T. )

    * update the vertical scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )

    * check to extend the bottom screen row by one cell grid
    IF ::nBotRowPtr + ::nGridHeight + ::nGridRowOff < ::nDataRowBot - (::nGridHeight + ::nGridRowOff) - IIF( ::nFootRowTop = -1, 0, ::nGridHeight )
      ::nBotRowPtr += ::nGridHeight + ::nGridRowOff
    ENDIF

    * release the display context to Windows
    ReleaseDC( ::hWnd, hDC )
    RETURN 0

  ENDIF

  * check to highlight the current cell or paint the entire line
  IF ::lCellPaint
    ::HiLight( hDC )
  ELSE
    ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
  ENDIF

  * update the vertical scrollbar
  ::oVScroll:SetPos( Eval( ::bRECPOS ) )

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD GoDown() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * skip forward and check for eof, exit if so
  Eval( ::bSKIPPER, 1 )
  IF Eval( ::bEOF )
    Eval( ::bGOBOTTOM )
    RETURN 0
  ENDIF

  * obtain our display context and font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * skip back and delight the previous row
  Eval( ::bSKIPPER, -1 )

  * check to delight the current cell or paint the entire line
  IF ::lCellPaint
    ::DeLight( hDC )
  ELSE
    ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
  ENDIF

  * skip to the new source data element and record it
  Eval( ::bSKIPPER, 1 )
  ::nDataCurElm := Eval( ::bRECNO )

  * advance the screen row pointer
  ::nCurRowPtr += ::nGridHeight + ::nGridRowOff

  * check to see if we need to scroll the data are upwards
  IF ::nCurRowPtr + ::nGridHeight + ::nGridRowOff > ::nDataRowBot

    * restore the previous screen row
    ::nCurRowPtr -= ::nGridHeight + ::nGridRowOff

    * advance and record the top data element 
    Eval( ::bGOTO, ::nDataTopElm )
    Eval( ::bSKIPPER, 1 )
    ::nDataTopElm := Eval( ::bRECNO )

    * advance and record the bottom data element 
    Eval( ::bGOTO, ::nDataBotElm )
    Eval( ::bSKIPPER, 1 )
    ::nDataBotElm := Eval( ::bRECNO )

    * reposition on the current data element
    Eval( ::bGOTO, ::nDataCurElm )

    * inform our paint method to keep out of the way
    ::lRePaint := .F.

    * scroll the data area down and repaint to satisfy windows
    ScrollWindow( ::hWnd, 0, -(::nGridHeight + ::nGridRowOff), ;
                   {  ::nDataRowTop, ::nTlc, ::nDataRowBot, ::nBrc }, ;
                   {  ::nDataRowTop, ::nTlc, ::nDataRowBot, ::nBrc } )
    ::BeginPaint()
    ::Paint()
    ::EndPaint()

    * paint the data line with grids
    ::LinePaint( hDC, ::nCurRowPtr, .T., .T. )

    * check to allow the display of one last cell beyond the bottom row
    IF ::nFootRowTop = -1

      * skip to the next data element and paint it
      Eval( ::bSKIPPER, 1 )
      SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
      SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )
      IF Eval( ::bEOF )
        ::BlankPaint( hDC, ::nCurRowPtr + ::nGridHeight + ::nGridRowOff, .T. )
      ELSE
        ::LinePaint( hDC, ::nCurRowPtr + ::nGridHeight + ::nGridRowOff, .T., .F. )
      ENDIF
      * reposition on the current data element
      Eval( ::bGOTO, ::nDataCurElm )

    ENDIF

    * update the vertical scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )

    * release the display context to Windows
    ReleaseDC( ::hWnd, hDC )
    RETURN 0

  ENDIF

  * check to highlight the current cell or paint the entire line
  IF ::lCellPaint
    ::HiLight( hDC )
  ELSE
    ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
  ENDIF

  * update the vertical scrollbar
  ::oVScroll:SetPos( Eval( ::bRECPOS ) )

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD GoPageUp() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * obtain our display context and font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * move to the top displayed data element and skip backward
  Eval( ::bGOTO, ::nDataTopElm )
  Eval( ::bSKIPPER, -1 )

  * check for bof
  IF Eval( ::bBOF )

    * go to the top of the source data
    Eval( ::bGOTOP )

    * check to see if we were already on the top data element
    IF ::nDataCurElm = Eval( ::bRECNO )
      * release the display context to Windows
      ReleaseDC( ::hWnd, hDC )
      RETURN 0
    ENDIF

    * delight the current cell
    Eval( ::bGOTO, ::nDataCurElm )

    * check to delight the current cell or paint the entire line
    IF ::lCellPaint
      ::DeLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
    ENDIF

    * set the top data element and row
    ::nDataCurElm := ::nDataTopElm
    ::nCurRowPtr := ::nDataRowTop

    * reposition on the current data element, highlight and return
    Eval( ::bGOTO, ::nDataCurElm )

    * check to highlight the current cell or paint the entire line
    IF ::lCellPaint
      ::HiLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
    ENDIF

    * update the vertical scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )

    * release the display context to Windows
    ReleaseDC( ::hWnd, hDC )

    RETURN 0

  ENDIF

  * skip backwards one screenfuls worth of source data elements
  Eval( ::bSKIPPER, ( Int( (::nDataRowBot - ::nDataRowTop ) / ( ::nGridHeight + ::nGridRowOff ) ) -1 ) * -1 )

  * check for bof and go top if so
  IF Eval( ::bBOF )
    Eval( ::bGOTOP )
  ENDIF

  * record the new top displayed data element and repaint
  ::nDataTopElm := Eval( ::bRECNO )
  ::DataPaint( hDC, .F. )

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD GoPageDown() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * obtain our display context and font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * move to the bottom displayed data element and skip forward
  Eval( ::bGOTO, ::nDataBotElm )
  Eval( ::bSKIPPER, 1 )

  * check for eof
  IF Eval( ::bEOF )

    * go to the bottom of the source data
    Eval( ::bGOBOTTOM )

    * check to see if we were already on the bottom data element
    IF ::nDataCurElm = Eval( ::bRECNO )
      * release the display context to Windows
      ReleaseDC( ::hWnd, hDC )
      RETURN 0
    ENDIF

    * delight the current cell
    Eval( ::bGOTO, ::nDataCurElm )

    * check to delight the current cell or paint the entire line
    IF ::lCellPaint
      ::DeLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
    ENDIF

    * move down to the end of data, advancing the current row as we go
    DO WHILE .T.
      Eval( ::bSKIPPER, 1 )
      IF Eval( ::bEOF )
        EXIT
      ENDIF
      ::nCurRowPtr += ::nGridHeight + ::nGridRowOff
    ENDDO

    * reposition on the bottom data element
    Eval( ::bGOBOTTOM )
    ::nDataCurElm := Eval( ::bRECNO )

    * check to highlight the current cell or paint the entire line
    IF ::lCellPaint
      ::HiLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
    ENDIF

    * update the vertical scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )

    * release the display context to Windows
    ReleaseDC( ::hWnd, hDC )

    RETURN 0

  ENDIF

  * record the new top displayed data element and repaint
  ::nDataTopElm := Eval( ::bRECNO )
  ::DataPaint( hDC, .F. )

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD GoTop() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * obtain our display context and font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * move to the top displayed data element and skip backward
  Eval( ::bGOTO, ::nDataTopElm )
  Eval( ::bSKIPPER, -1 )

  * check for bof
  IF Eval( ::bBOF )
    * go to the top of the source data
    Eval( ::bGOTOP )

    * check to see if we were already on the top data element
    IF ::nDataCurElm = Eval( ::bRECNO )
      * release the display context to Windows
      ReleaseDC( ::hWnd, hDC )
      RETURN 0
    ENDIF

    * delight the current cell
    Eval( ::bGOTO, ::nDataCurElm )

    * check to delight the current cell or paint the entire line
    IF ::lCellPaint
      ::DeLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
    ENDIF

    * set the top data element and row
    ::nDataCurElm := ::nDataTopElm
    ::nCurRowPtr := ::nDataRowTop

    * reposition on the current data element, highlight and return
    Eval( ::bGOTO, ::nDataCurElm )

    * check to highlight the current cell or paint the entire line
    IF ::lCellPaint
      ::HiLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
    ENDIF

    * update the vertical scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )

    * release the display context to Windows
    ReleaseDC( ::hWnd, hDC )

    RETURN 0

  ENDIF

  * delight the current cell
  Eval( ::bGOTO, ::nDataCurElm )

  * check to delight the current cell or paint the entire line
  IF ::lCellPaint
    ::DeLight( hDC )
  ELSE
    ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
  ENDIF

  * go to the top of the source data
  Eval( ::bGOTOP )

  * record the new top displayed data element and repaint
  ::nDataTopElm := Eval( ::bRECNO )
  ::nDataCurElm := ::nDataTopElm
  ::nCurRowPtr := ::nDataRowTop
  ::DataPaint( hDC )

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD GoBottom() CLASS KbDialog
******************************************************************************
  LOCAL hDC

  * obtain our display context and font
  hDC := GetDC( ::hWnd )
  SelectObject( hDC, ::oFont:hFont )

  * move to the bottom displayed data element and skip forward
  Eval( ::bGOTO, ::nDataBotElm )
  Eval( ::bSKIPPER, 1 )

  * check for eof
  IF Eval( ::bEOF )

    * go to the bottom of the source data
    Eval( ::bGOBOTTOM )

    * check to see if we were already on the bottom data element
    IF ::nDataCurElm = Eval( ::bRECNO )
      * release the display context to Windows
      ReleaseDC( ::hWnd, hDC )
      RETURN 0
    ENDIF

    * delight the current cell
    Eval( ::bGOTO, ::nDataCurElm )

    * check to delight the current cell or paint the entire line
    IF ::lCellPaint
      ::DeLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
    ENDIF

    * move down to the end of data, advancing the current row
    DO WHILE .T.
      Eval( ::bSKIPPER, 1 )
      IF Eval( ::bEOF )
        EXIT
      ENDIF
      ::nCurRowPtr += ::nGridHeight + ::nGridRowOff
    ENDDO

    * reposition on the bottom data element
    Eval( ::bGOBOTTOM )
    ::nDataCurElm := Eval( ::bRECNO )

    * check to highlight the current cell or paint the entire line
    IF ::lCellPaint
      ::HiLight( hDC )
    ELSE
      ::LinePaint( hDC, ::nCurRowPtr, .F., .T. )
    ENDIF

    * update the vertical scrollbar
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )

    * release the display context to Windows
    ReleaseDC( ::hWnd, hDC )

    RETURN 0

  ENDIF

  * delight the current cell
  Eval( ::bGOTO, ::nDataCurElm )

  * check to delight the current cell or paint the entire line
  IF ::lCellPaint
    ::DeLight( hDC )
  ELSE
    ::LinePaint( hDC, ::nCurRowPtr, .F., .F. )
  ENDIF

  * go to the bottom of the source data
  Eval( ::bGOBOTTOM )

  * skip backwards one screenfuls worth of source data elements
  Eval( ::bSKIPPER, ( Int( (::nDataRowBot - ::nDataRowTop ) / ( ::nGridHeight + ::nGridRowOff ) ) -1 ) * -1 )

  * record the new top displayed data element and repaint
  ::nDataTopElm := Eval( ::bRECNO )
  ::nCurRowPtr := ::nDataRowBot
  ::DataPaint( hDC )

  * release the display context to Windows
  ReleaseDC( ::hWnd, hDC )

RETURN 0

******************************************************************************
METHOD LButtonDown( nRow, nCol, nKeyFlags ) CLASS KBrowse
******************************************************************************
  LOCAL nRowSkip
  LOCAL nColPtr

  * optain input focus
  ::SetFocus()

  * check to see if the mouse is within our data area, return if not
  IF nRow < ::nDataRowTop .OR. nRow > ::nBotRowPtr + ::nGridHeight + ::nGridRowOff .OR. ;
    nCol < ::nTlc .OR. nCol > ::nBrc
    RETURN 0
  ENDIF

  * check for a valid data element
  IF ::nDataCurElm = 0
    RETURN 0
  ENDIF

  * determine the number of rows to skip
  nRowSkip := 0
  IF nRow < ::nCurRowPtr
    nRowSkip := ( Int( (::nCurRowPtr - nRow) / (::nGridHeight + ::nGridRowOff) ) + 1 ) * -1
  ENDIF
  IF nRow > ::nCurRowPtr + ::nGridHeight + ::nGridRowOff
    nRowSkip := Int( (nRow - ::nCurRowPtr) / (::nGridHeight + ::nGridRowOff) ) 
  ENDIF

  * determine the column the mouse cursor is in
  FOR nColPtr = IIF( ::nFrzColPtr = 0, ::nTopColPtr, 1 ) TO ::nBotColPtr
    IF nCol >= ::aColumns[ nColPtr ]:nScreenCol .AND. ;
       nCol <= ::aColumns[ nColPtr ]:nScreenCol + ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
      EXIT
    ENDIF
    IF nColPtr = ::nFrzColPtr
      nColPtr = ::nTopColPtr - 1
    ENDIF
  NEXT
  IF nColPtr > LEN( ::aColumns )
    nColPtr := LEN( ::aColumns )
  ENDIF

  * check to see if the mouse is in the highlighted cell, return if so
  IF nRowSkip = 0 .AND. nColPtr = ::nCurColPtr
    RETURN 0
  ENDIF

  * delight the current cell
  ::DeLight()

  * check to see if we need to skip any rows
  IF nRowSkip <> 0
    Eval( ::bSKIPPER, nRowSkip )
    ::nDataCurElm := Eval( ::bRECNO )
    ::nCurRowPtr += ((::nGridHeight + ::nGridRowOff) * nRowSkip)
    IF ::nCurRowPtr < ::nDataRowTop
      ::nCurRowPtr := ::nDataRowTop
    ENDIF
    IF ::nCurRowPtr > ::nBotRowPtr
      ::nCurRowPtr := ::nBotRowPtr
    ENDIF
    ::oVScroll:SetPos( Eval( ::bRECPOS ) )
  ENDIF   

  * check to see if we need to move the column pointer
  IF nColPtr <> ::nCurColPtr
    ::nCurColPtr := nColPtr
    ::oHScroll:SetPos( ::nCurColPtr )

    * check to see if the user highlighted the rightmost partial column
    IF ::nCurColPtr > ::nBotColPtr
      * adjust the top column element pointer until the new column moves onscreen
      nCol := ::aColumns[ ::nBotColPtr ]:nScreenCol + ;
              ::aColumns[ ::nBotColPtr ]:nCellWidth + ::nGridColOff + ;
              ::aColumns[ ::nCurColPtr ]:nCellWidth + ::nGridColOff
    
      FOR nColPtr := ::nTopColPtr TO ::nCurColPtr
        nCol -= ::aColumns[ nColPtr ]:nCellWidth + ::nGridColOff
        IF nCol <= ::nBrc
          nColPtr++
          EXIT
        ENDIF
      NEXT
      IF nColPtr > LEN( ::aColumns )
        nColPtr = LEN( ::aColumns )
      ENDIF
      ::nTopColPtr := nColPtr
      ::lClearPaint := .T.
      ::BeginPaint()
      ::Paint()
      ::EndPaint()
      RETURN 0
    ENDIF
  ENDIF
  ::HiLight()

RETURN 0

******************************************************************************
METHOD VScroll( nWParam, nLParam ) CLASS KbDialog
******************************************************************************

  * check for a valid data element
  IF ::nDataCurElm = 0
    RETURN 0
  ENDIF

  DO CASE

  CASE nWParam == SB_LINEUP
    RETURN ::GoUp()

  CASE nWParam == SB_LINEDOWN
    RETURN ::GoDown()

  CASE nWParam == SB_PAGEUP
    RETURN ::GoUp()

  CASE nWParam == SB_PAGEDOWN
    RETURN ::GoDown()

  CASE nWParam == SB_TOP
    RETURN ::GoTop()

  CASE nWParam == SB_BOTTOM
    RETURN ::GoBottom()

  CASE nWParam == SB_THUMBPOSITION
    Eval( ::bGOTOPOS, nLParam )
    IF ::aReference = NIL
      IF Deleted() .AND. SET( 11 )
        * skip forward and check for eof, exit if so
        Eval( ::bSKIPPER, 1 )
        IF Eval( ::bEOF )
          Eval( ::bGOBOTTOM )
          IF Eval( ::bEOF )
            RETURN 0
          ENDIF
        ENDIF
      ENDIF
    ENDIF
    ::nDataTopElm := Eval( ::bRECNO )
    ::nDataCurElm := Eval( ::bRECNO )
    ::BeginPaint()
    ::Paint()
    ::EndPaint()

  ENDCASE

RETURN 0

******************************************************************************
METHOD HScroll( nWParam, nLParam ) CLASS KbDialog
******************************************************************************

  DO CASE

  CASE nWParam == SB_LINEUP
    RETURN ::GoLeft()

  CASE nWParam == SB_LINEDOWN
    RETURN ::GoRight()

  CASE nWParam == SB_PAGEUP
    RETURN ::GoLeft()

  CASE nWParam == SB_PAGEDOWN
    RETURN ::GoRight()

  CASE nWParam == SB_TOP
    RETURN ::GoFarLeft()

  CASE nWParam == SB_BOTTOM
    RETURN ::GoFarRight()

  CASE nWParam == SB_THUMBPOSITION
    ::nCurColPtr := nLParam
    ::nTopColPtr := nLParam
    IF ::nFrzColPtr > 0 .AND. ::nCurColPtr <= ::nFrzColPtr
      ::nTopColPtr := ::nFrzColPtr + 1
    ENDIF
    ::oHScroll:SetPos( ::nCurColPtr )
    ::lClearPaint := .T.
    ::BeginPaint()
    ::Paint()
    ::EndPaint()

  ENDCASE

RETURN 0

******************************************************************************
METHOD SetNewFont( oFont ) CLASS KbDialog
******************************************************************************

  IF oFont = NIL
    DEFINE FONT oFont FROM USER
  ENDIF

  IF .NOT. EMPTY( oFont:cFaceName )
    ::oFont := oFont
    ::lCalcFont := .T.
    ::lCalcSize := .T.
    ::lClearPaint := .T.
    ::nCurColPtr := ::nTopColPtr
    ::BeginPaint()
    ::Paint()
    ::EndPaint()
  ENDIF
  ::SetFocus()

RETURN 0

******************************************************************************
METHOD SetFreeze( nColPtr ) CLASS KbDialog
******************************************************************************

  ::nFrzColPtr := nColPtr
  ::nTopColPtr := nColPtr + 1
  ::nCurColPtr := nColPtr + 1
  ::nBotColPtr := nColPtr + 1

RETURN NIL

******************************************************************************
METHOD SetGrid( nGridType ) CLASS KbDialog
******************************************************************************

  * check for types 1,2 or 3
  IF nGridType = NIL .OR. nGridType < 1 .OR. nGridType > 3 
    nGridType := 3
  ENDIF
  ::nGridType := nGridType

  DO CASE
  CASE ::nGridType = 1  
    ::nGridRowAdj := 1  // adjust offsets for single line boxes
    ::nGridColAdj := 1
    ::nGridRowOff := 0
    ::nGridColOff := 0
    ::nTextColAdj := 2
    ::nTextBoxPix := 1
  CASE ::nGridType = 2 
    ::nGridRowAdj := 1  // adjust offsets for double line boxes
    ::nGridColAdj := 1
    ::nGridRowOff := 2
    ::nGridColOff := 2
    ::nTextColAdj := 2
    ::nTextBoxPix := 1
  CASE ::nGridType = 3
    ::nGridRowOff := 2  // adjust offsets for 3-dimensional boxes
    ::nGridColOff := 2  
    ::nGridRowAdj := 2  
    ::nGridColAdj := 2  
    ::nTextColAdj := 2
    ::nTextBoxPix := 0
  ENDCASE
  ::lCalcFont := .T.
  ::lCalcSize := .T.
  ::lClearPaint := .T.

RETURN 0

******************************************************************************
METHOD SetHeadColor( nFgColor, nBgColor ) CLASS KbDialog
******************************************************************************

  ::aColors[ KTB_HEADCLR ][ KTB_FOREGROUND ] := nFgColor
  ::aColors[ KTB_HEADCLR ][ KTB_BACKGROUND ] := nFgColor

RETURN 0

******************************************************************************
METHOD SetDataColor( nFgColor, nBgColor ) CLASS KbDialog
******************************************************************************

  ::aColors[ KTB_DATACLR ][ KTB_FOREGROUND ] := nFgColor
  ::aColors[ KTB_DATACLR ][ KTB_BACKGROUND ] := nFgColor

RETURN 0

******************************************************************************
METHOD SetFootColor( nFgColor, nBgColor ) CLASS KbDialog
******************************************************************************

  ::aColors[ KTB_FOOTCLR ][ KTB_FOREGROUND ] := nFgColor
  ::aColors[ KTB_FOOTCLR ][ KTB_BACKGROUND ] := nFgColor

RETURN 0

******************************************************************************
METHOD SetLineColor( nFgColor, nBgColor ) CLASS KbDialog
******************************************************************************

  ::aColors[ KTB_LINECLR ][ KTB_FOREGROUND ] := nFgColor
  ::aColors[ KTB_LINECLR ][ KTB_BACKGROUND ] := nFgColor

RETURN 0

******************************************************************************
METHOD SetCellColor( nFgColor, nBgColor ) CLASS KbDialog
******************************************************************************

  ::aColors[ KTB_CELLCLR ][ KTB_FOREGROUND ] := nFgColor
  ::aColors[ KTB_CELLCLR ][ KTB_BACKGROUND ] := nFgColor

RETURN 0

******************************************************************************
* this code fragment is used to print a message at 0,0 for debugging purposes
******************************************************************************
*     hDC := GetDC( ::hWnd )
*     SelectObject( hDC, ::oFont:hFont )
*     SetTextColor( hDC, ::aColors[ KTB_DATACLR, KTB_FOREGROUND ] )
*     SetBkColor( hDC, ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )
*     EXTTEXTOUT( hDC, ::nTlr, ::nTlc, { ::nTlr, ::nTlc, ::nTlr + 20, ::nTlc + 100 }, "MADE IT HERE" )
*     ReleaseDC( ::hWnd, hDC )
******************************************************************************
