******************************************************************************
* PROGRAM ----: KBMCHILD.PRG 
* AUTHOR -----: Kevin E. Saffer 
* COPYRIGHT --: None, placed into the public domain
* CREATED ----: 6/12/1994 at 17:50
******************************************************************************
#include "fivewin.ch"

* offsets into the browse color array aColors
#define KTB_HEADCLR      1  // header area color pair
#define KTB_DATACLR      2  // data area color pair
#define KTB_FOOTCLR      3  // footer area color pair
#define KTB_LINECLR      4  // line color pair
#define KTB_CELLCLR      5  // cell pointer color pair

* offsets into each browse color array element
#define KTB_FOREGROUND   1
#define KTB_BACKGROUND   2

* additional defines not in fivewin.ch
#define WM_MDICREATE        544   // 0x0220
#define WM_MDIDESTROY       545   // 0x0221
#define WM_MDIACTIVATE      546   // 0x0222
#define WM_MDIRESTORE       547   // 0x0223
#define WM_MDINEXT          548   // 0x0224
#define WM_MDIMAXIMIZE      549   // 0x0225
#define WM_MDITILE          550   // 0x0226
#define WM_MDICASCADE       551   // 0x0227
#define WM_MDIICONARRANGE   552   // 0x0228
#define WM_MDIGETACTIVE     553   // 0x0229
#define WM_MDISETMENU       554   // 0x0230

#define WM_CHILDACTIVATE     34   // 0x22
#define WM_GETFONT           49   // 0x0031

#define SC_RESTORE        61728
#define SC_CLOSE          61536   // 0xF060
#define SC_MINIMIZE       61472
#define SC_NEXT           61504
#define SC_MAXIMIZE       61488   // 0xF030

#define IDC_ARROW         32512
#define WM_SYSCOMMAND       274   // 0x0112
#define CW_USEDEFAULT     32768
#define CS_DBLCLKS            8

STATIC lRegistered := .F.

******************************************************************************
* class definition, instance variables, and method prototypes
******************************************************************************
CLASS KbMChild FROM KbWindow

  DATA   oWndClient, oControl

  METHOD New() CONSTRUCTOR

  METHOD KeyDown()

  METHOD Maximize() INLINE ::oWndClient:ChildMaximize( Self )

  METHOD Restore()  INLINE ::oWndClient:ChildRestore( Self )

  METHOD Activate( cShow, bLClicked, bRClicked, bMoved, bResized, bPainted,;
                    bKeyDown, bInit,;
                    bUp, bDown, bPgUp, bPgDn,;
                    bLeft, bRight, bPgLeft, bPgRight, bValid ) INLINE ;
                    ;
                    ::oWndClient:ChildActivate( Self ),;
                    Super:Activate( cShow, bLClicked, bRClicked, bMoved,;
                    bResized, bPainted, bKeyDown, bInit,;
                    bUp, bDown, bPgUp, bPgDn,;
                    bLeft, bRight, bPgLeft, bPgRight, bValid ),;
                    ::lVisible := .t.,;
                    If( ::oControl != nil, ::oControl:SetFocus(),)

  METHOD End()

  METHOD GotFocus()

  METHOD Next() INLINE ::oWndClient:ChildNext( Self )

  METHOD Prev() INLINE ::oWndClient:ChildNext( Self, .t. )

  METHOD ReSize() BLOCK { | Self, oRect | ;     // It has to be BLOCK
                   Super:ReSize(),;
                   oRect := ::GetCliRect(),;
                   oRect:nTop += If( ::oBar != nil, ::oBar:nBtnHeight, 0 ),;
                   oRect:nLeft--,;
                   If( ::oControl != nil, ::oControl:SetCoors( oRect ),), nil }

  METHOD SetControl( oCtrl ) INLINE ::oControl := oCtrl, ::ReSize()

ENDCLASS

******************************************************************************
METHOD New( oWnd, nTlr, nTlc, nBrr, nBrc, cCaption, oFont ) CLASS KbMChild
******************************************************************************

  DEFAULT oWnd := GetWndFrame(),;
          nTlr := oWnd:nTop, ;
          nTlc := oWnd:nLeft, ;
          nBrr := oWnd:nBottom, ;
          nBrc := oWnd:nRight, ;
          cCaption := "", ;
          oFont := TFont():New( "Fixedsys", 8, 12 )

  ::nTop := nTlr
  ::nLeft := nTlc
  ::nBottom := nBrr
  ::nRight := nBrc
  ::oWndClient := oWnd:oWndClient
  ::oWnd := oWnd
  ::lVisible := .T.
  ::aControls := {}
  ::oFont := oFont
  ::lFocused := .T.

  ::aColors := { { CLR_WHITE, CLR_BLUE }, ;
                 { CLR_BLACK, CLR_LIGHTGRAY }, ;
                 { CLR_WHITE, CLR_GRAY }, ;
                 { CLR_BLACK, CLR_LIGHTGRAY }, ;
                 { CLR_WHITE, CLR_BLACK } ;
               }

  ::aColumns := {} 
  ::aReference := NIL

  ::SetColor( ::aColors[ KTB_DATACLR, KTB_FOREGROUND ], ::aColors[ KTB_DATACLR, KTB_BACKGROUND ] )

  * initialize the column array pointers
  ::nFrzColPtr := 0
  ::nTopColPtr := 1
  ::nCurColPtr := 1
  ::nBotColPtr := 1

  * default to grid type 3 (3D Boxes)
  ::SetGrid( 3 )

  * initialize the screen and data pointers
  ::nHeadOffset := 0
  ::nFootOffset := 0
  ::nCurRowPtr := 0
  ::nBotRowPtr := 0 
  ::nDataTopElm := 0
  ::nDataCurElm := 0
  ::nDataBotElm := 0
  ::nArrCurElm := 1

  * set up lastkey
  ::nLastKey := 0

  * create standard database file operation blocks
  ::bBOF := { || BOF() }
  ::bEOF := { || EOF() }
  ::bGOTOP  := { || DBGOTOP() }
  ::bGOBOTTOM := { || DBGOBOTTOM() }
  ::bGOTO := { | n | DBGOTO( n ) }
  ::bSKIPPER := { | n | DBSKIP( n ) }
  ::bRECNO := { || RECNO() }
  ::bRECCOUNT := { || RECCOUNT() }
  ::bRECPOS := { || IIF( INDEXORD() = 0, RECNO(), NTXPOS( INDEXORD(), RECNO() ) ) }
  ::bGOTOPOS := { | n | DBGOTO( n ) }

  * store the database alias and indexorder
  ::cAlias := ALIAS()
  ::nWorkArea := SELECT()
  ::nIndexOrd := INDEXORD()

  * create the scrollbars
  DEFINE SCROLLBAR ::oVScroll VERTICAL OF Self
  DEFINE SCROLLBAR ::oHScroll HORIZONTAL OF Self

  IF ! lRegistered
    RegisterClass( "TMDICHILD", nOR( CS_VREDRAW, CS_HREDRAW, CS_DBLCLKS), 0, 0, 0, 0, 0, "", GetMDIChlProc() )
    lRegistered = .T.
  ENDIF

  ::hWnd := ::oWndClient:SendMsg( WM_MDICREATE, 0, cMdiStruct( "TMDICHILD", cCaption, ::nTop, ::nLeft, ::nBottom, ::nRight, ::nStyle ) )
  ::Default()
  ::Link()
  ::oWndClient:Add( Self )

  ::SetFocus()

  SetWndDefault( Self )

RETURN 0

******************************************************************************
METHOD GotFocus() CLASS KbMChild
******************************************************************************

  IF ::oControl != NIL
    ::oControl:SetFocus()
  ENDIF

RETURN Super:GotFocus()

******************************************************************************
METHOD KeyDown( nKey, nFlags ) CLASS KbMChild
******************************************************************************

  IF nKey == VK_F4 .and. GetKeyState( VK_CONTROL )
    ::SendMsg( WM_SYSCOMMAND, SC_CLOSE )
    RETURN 0
  ENDIF

  IF nKey == VK_F6 .and. GetKeyState( VK_CONTROL )
    ::SendMsg( WM_SYSCOMMAND, SC_NEXT )
    RETURN 0
  ENDIF

RETURN Super:KeyDown( nKey, nFlags )

******************************************************************************
METHOD End() CLASS KbMChild
******************************************************************************
  LOCAL lEnd := .T.

  IF ::bValid != NIL
    IF Eval( ::bValid )
      ::oWndClient:ChildClose( Self )
    ELSE
      lEnd = .F.
    ENDIF
  ELSE
    ::oWndClient:ChildClose( Self )
  ENDIF

RETURN lEnd

